//---------------------------------------------------------------------------
// File  :  fft_filter_demo_1.c
// Date  :  2012-02-19   (YYYY-MM-DD)
// Author:  Wolfgang Buescher  (DL4YHF)
// Location: <DL4YHF>C:\cbproj\SpecLab\fft_filter_plugins\*.*
//
// Description:
//   Sample plugin for Spectrum Lab's FFT-based filter .
//   Processes signals in the FREQUENCY DOMAIN (not in the time domain).
//   Compile with DevC++ (! - NOT "wxDevC++!!!) or Borland C++ Builder V4;
//           MS-C may work but was never tested.
//   Beware: Avoid "wxDevCpp"- It hijacks existing DevC++ projects,
//           and turns on the stupid name mangling again, i.e. produces
//           function names like "FFP_Init@4" instead of "FFP_Init" in the DLL
//           rendering it completely useless.
//           Even worse, installing wxDevC++ seems to overwrite Dev-C++ V4.9.9.2,
//           so you won't be able to compile the demo plugin with Dev-C++
//           after installing wxDev-C++ (which, besides this annoyance,
//           is a nice free C++ development system).
//
// Latest Changes:
//   2012-02-19 : Added a test for the timestamps in FFP_ProcessTimeDomain().
//   2009-09-27 : Added FFP_ProcessDisplaySpectrum() in the demo plugin.
//   2007-03-11 : Wrote the first, simple example of a filter plugin.

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "fft_filter_dll.h" /* header for all FFT-filter plugin DLLs    */
#include "FFP_framework.h"  /* some framework for the FFT-filter plugin */

//---------------------------------------------------------------------
// Compilation switches
//---------------------------------------------------------------------

#ifndef  SWI_PLUGIN_WITH_GUI  /* shall the plugin have its own control panel ? */
 #define SWI_PLUGIN_WITH_GUI 1   /* 0=no, 1=yes */
#endif

#ifdef __BORLANDC__
 #pragma argsused   // Borland gripes "never used" - so what  ;-)
#endif


//---------------------------------------------------------------------
// Internal defines
//---------------------------------------------------------------------

#if(SWI_PLUGIN_WITH_GUI)  // If the plugin has it's own "GUI" (control panel):
 // Define a *unique*  class name for the plugin's own window !
 // C_PLUGIN_WINDOW_CLASS = name of the window class; max. 40 characters .
 // To avoid conflicts with the "window class name" when running two plugins
 // compiled with different compilers at the same time, we use different names
 // for the plugin, depending on the development environment.
 // Without this, RegisterClassEx() would fail (see FFP_frameworkX.c) .
 #ifdef __BORLANDC__     /* Borland C, Borland C++Builder : */
   #define C_PLUGIN_WINDOW_CLASS "FFT Filter Plugin 1 BC"
 #endif
 #ifdef __LCC__          /* lcc C compiler system for Win32: */
   #define C_PLUGIN_WINDOW_CLASS "FFT Filter Plugin 1 LCC"
 #endif
 #ifdef __GNUC__         /* GNU C compiler, including MinGW or DevCpp */
   #define C_PLUGIN_WINDOW_CLASS "FFT Filter Plugin 1 GCC"
 #endif
 #ifndef C_PLUGIN_WINDOW_CLASS
  #define C_PLUGIN_WINDOW_CLASS "FFT Filter Plugin 1" /* unique name, max. 40 chars */
 #endif
#endif // (SWI_PLUGIN_WITH_GUI)

//---------------------------------------------------------------------
// Internal data types
//---------------------------------------------------------------------

typedef struct // extra filter info / internal values for each instance of this plugin.
{ //  A pointer to this structure 'extends' the
  //  T_FftFilterInfo structure (.pExtra) for everything
  //  which we need in the processing callbacks,
  //  but don't want to have in a global variable .
  //  It also simplifies the job to support multiple instances
  //  of this plugin running "in parallel" .
  char   sz80WindowTitle[84]; // text which shall appear in the plugin's window title
  double dblPhasePlotRe; // real part for phase-plot, normalized to +/- 1.0
  double dblPhasePlotIm; // imaginary part "  " " ..
  int    old_phase_x, old_phase_y;  // to erase old phase plot
  BOOL   fDoPhasePlot;   // flag from audio thread for plotter
  double dblNcoPhase;    // current phase (angle) of the oscillator
} T_MyInstData;
#define THIS_INST ((T_MyInstData*)pInfo->pInstData)

//---------------------------------------------------------------------
// Internal variables
//---------------------------------------------------------------------

  /* none ! */



//---------------------------------------------------------------------
// Internal functions
//---------------------------------------------------------------------

/***************************************************************************/
void FFP_PrintErrorString( T_FftFilterInfo *pInfo, char *pszFormat, ... )
  /* Prints a string into the multiline message window
   *  on the filter Control Panel (inside SpecLab).
   *  Caution: don't call too often, because SL only polls a flag
   *  every 100 milliseconds (approx).
   */
{
  va_list parameter;
  va_start( parameter, pszFormat/*lastfix*/ );
  wvsprintf( pInfo->sz255ErrorText, pszFormat, parameter );
  va_end(parameter);
  ++pInfo->iErrorTextModifyCnt; // acts as a "signal" for the filter c.p.
} // end FFP_PrintErrorString()


/***************************************************************************/
double CalculateAngle(double re, double im)
  // Four-quadrant conversion of a complex pair ("I/Q")
  //   into an phase value (in radians, but explained in degrees here).
  // A positive real value gives an angle of zero, etc.
  // Returned value range is -180 .. +180 (but in radians) =  -pi .. pi .
  // If both real and imaginary part are zero, the returned value
  // is zero.
  // Revision history:
  //   Jan 13, 2002, by DL4YHF:
  //       Implemented for a phase-sensitive spectrum analyzer
  //   March 15, 2002:
  //       Copied into the sound thread module
  //   Sept  15, 2002:
  //       Copied into AM_FM_DeMod.cpp (sound utilities)
  //   2007-03-11 :
  //       Copied into the FFT-filter-plugin demo .
{
#define ANGLE_RANGE_PLUS_MINUS_180_DEGREES 1  // 1: result_range = -pi..pi = -180..+180
     // (-180..+180 is preferred because angles tend to be +-0.x degrees,
     //  and it looks ugly if the display jumps from "0.1" to "359.9" and back)
  if(im > 0.0)
   {  // first or second quadrant
     if( re > 0.0 )
      { // first quadrant (0..90 degees)
        return atan(im/re);
      }
     else
     if( re < 0.0 )
      { // second quadrant (90..180 degrees)
        return atan(im/re) + C_PI;
      }
     else // re=0, im>0
      {
        return 0.5 * C_PI;
      }
   }
  else // ! im>0
  if(im < 0.0)
   {  // third or fourth quadrant
     if( re < 0.0 )
      { // third quadrant
#if(ANGLE_RANGE_PLUS_MINUS_180_DEGREES)
        return atan(im/re) - C_PI;     // for result range -180..-90
#else
        return atan(im/re) + C_PI;     // for result range 180..270
#endif
      }
     else
     if( re > 0.0 )
      { // fourth quadrant
#if(ANGLE_RANGE_PLUS_MINUS_180_DEGREES)
        return atan(im/re);           // for result range -90..0
#else
        return atan(im/re) + 2*C_PI;  // for result range 270..360
#endif
      }
     else // re=0, im<0  -> 270 degrees
      {
#if(ANGLE_RANGE_PLUS_MINUS_180_DEGREES)
        return -0.5 * C_PI;
#else
        return 1.5 *  C_PI;
#endif
      }
   }
  else   // im=0, a "real" number
   {
     if(re>=0)
        return 0;
     else
        return C_PI;    // negative -> 180 degrees
   }
} // end ..CalculateAngle()


/***************************************************************************/
int FFP_FrequencyToBinIndex( T_FftFilterInfo *pInfo, double dblFreq_Hz )
  /* Converts an audio frequency (in Hertz) into an array index for the spectrum.
   */
{
  int iBinIndex;
  if( pInfo->dblFftBinWidth > 0.0 )
   { iBinIndex = (int)((dblFreq_Hz - pInfo->dblAudioFreqMin) / pInfo->dblFftBinWidth);
     if( iBinIndex<0 )
         iBinIndex=0;   // should never happen.. but safety first !
     if( iBinIndex>=pInfo->iNumFrequencyBins )
         iBinIndex= pInfo->iNumFrequencyBins-1;
   }
  else
   { iBinIndex = 0;
   }
  return iBinIndex;

} // end FFP_FrequencyToBinIndex()

/***************************************************************************/
double FFP_BinIndexToFrequency( T_FftFilterInfo *pInfo, int iBinIndex )
  /* Converts an array index for the spectrum into an audio frequency (in Hertz).
   *  (strictly spoken, it calculates the CENTER FREQUENCY for a certain
   *   FFT frequency bin ).
   * Inverse function to FFP_FrequencyToBinIndex() .
   */
{ return pInfo->dblAudioFreqMin + (double)iBinIndex * pInfo->dblFftBinWidth;
} // end FFP_BinIndexToFrequency()


/***************************************************************************/
void FFP_ShiftFrequencies(  // .. in the frequency domain (!)
            T_FftFilterInfo *pInfo,
            float *pfltFreqBinsRe,
            float *pfltFreqBinsIm,
            int iStartBinIndex,
            int iNumFrequBins,
            int iShiftOffset )
{
  int iAbsOffset = (iShiftOffset>0) ? iShiftOffset : -iShiftOffset;
  int iSrcIndex, iDstIndex;
  float fltTemp = 1.0; // inversion factor to compensate 180-problem (see DF6NM's note)

  // The original code only worked for an EVEN number of shifts. So:
  if( (iAbsOffset & 1)!=0 && (pInfo->dwFftCounter & 1)!=0)
   { // odd bin-offset, and odd FFT-counter ?
     fltTemp = -1.0;  // -> invert every second FFT (*)
   }
  // (*) A nice explanation for this from Markus (DF6NM) :
  // Eine Verschiebung um ein Bin bedeutet doch, dass die Phase genau 1 mal 360
  // pro FFT-Lnge umluft. Also bekommt ein Sample in der Mitte
  // der zweiten Hlfte 180 Differenz im Vergleich zur Mitte der
  // ersten Hlfte. Die beiden cos^2-Flanken, aus denen das Signal
  // wieder zusammengesetzt werden soll, stammen aber gerade aus zwei
  // um eine _halbe_ Blocklnge versetzten FFT's - also muss es
  // in der Mitte der Flanke eine Auslschung geben.
  // Abhilfe: fr ungerade Verschiebungen einfach jede zweite FFT um 180 drehen.
  if( iShiftOffset > 0 )
   { // moving UP, and clear bins[0..iShiftOffset-1] :
     // source index < destination index, so begin with the highest index:
     iSrcIndex = iStartBinIndex;
     for( iDstIndex=iNumFrequBins-1; iDstIndex>=0; --iDstIndex )
      { iDstIndex = iSrcIndex+iAbsOffset;
        if(iSrcIndex>0 && iSrcIndex<iNumFrequBins )
         { pfltFreqBinsRe[iDstIndex] = fltTemp * pfltFreqBinsRe[iSrcIndex];
           pfltFreqBinsIm[iDstIndex] = fltTemp * pfltFreqBinsIm[iSrcIndex];
         }
        else
         { pfltFreqBinsRe[iDstIndex] = 0.0;
           pfltFreqBinsIm[iDstIndex] = 0.0;
         }
      }
   } // end if < move frequencies UP >
  else // move down, and clear bins[iNumFrequBins-iAbsOffset..iNumFrequBins-1]
   { // source index will be greater than destination index, so start at the lowest index:
     for( iDstIndex=0; iDstIndex<iNumFrequBins; ++iDstIndex )
      { iSrcIndex = iDstIndex+iAbsOffset;
        if(iSrcIndex>0 && iSrcIndex<iNumFrequBins )
         { pfltFreqBinsRe[iDstIndex] = fltTemp * pfltFreqBinsRe[iSrcIndex];
           pfltFreqBinsIm[iDstIndex] = fltTemp * pfltFreqBinsIm[iSrcIndex];
         }
        else
         { pfltFreqBinsRe[iDstIndex] = 0.0;
           pfltFreqBinsIm[iDstIndex] = 0.0;
         }
      }
   }
} // end CFftFilter::ShiftFrequencies()

//---------------------------------------------------------------------
void FFP_InitParams(T_FftFilterInfo *pInfo, BOOL fRestoreDefaults )
  // Initializes the plugin's "parameters", and
  //  optionally sets everything (pInfo->params[].dblValue) to defaults.
  // Called when initializing, or via command from the debugging console.
{

  #define PN_SOURCE_START 0
  strcpy(pInfo->params[PN_SOURCE_START].sz40Name, "Source Start [Hz]" );
  pInfo->params[PN_SOURCE_START].dblMinValue = pInfo->dblAudioFreqMin; // 0 or -fs/2
  pInfo->params[PN_SOURCE_START].dblMaxValue = pInfo->dblSampleRate / 2.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_SOURCE_START].dblValue = 1000.0;   // default value, may be modified later
   }
  #define PN_RANGE_WIDTH  1
  strcpy(pInfo->params[PN_RANGE_WIDTH].sz40Name, "Range Width [Hz]" );
  pInfo->params[PN_RANGE_WIDTH].dblMinValue = 0.0;
  pInfo->params[PN_RANGE_WIDTH].dblMaxValue = pInfo->dblSampleRate / 2.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_RANGE_WIDTH].dblValue = 2000.0;
   }
  #define PN_DESTINATION_START 2
  strcpy(pInfo->params[PN_DESTINATION_START].sz40Name, "Destination Start [Hz]" );
  pInfo->params[PN_DESTINATION_START].dblMinValue = pInfo->dblAudioFreqMin; // 0 or -fs/2
  pInfo->params[PN_DESTINATION_START].dblMaxValue = pInfo->dblSampleRate / 2.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_DESTINATION_START].dblValue = 3000.0;
   }
  #define PN_GAIN_FACTOR  3
  strcpy(pInfo->params[PN_GAIN_FACTOR].sz40Name, "Gain (factor)" );
  pInfo->params[PN_GAIN_FACTOR].dblMinValue = -1.0;
  pInfo->params[PN_GAIN_FACTOR].dblMaxValue = +1.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_GAIN_FACTOR].dblValue = -1.0;
   }
  #define PN_MARKER1_FREQ 4
  strcpy(pInfo->params[PN_MARKER1_FREQ].sz40Name, "Marker 1 frequency [Hz]" );
  pInfo->params[PN_MARKER1_FREQ].dblMinValue = pInfo->dblAudioFreqMin; // 0 or -fs/2
  pInfo->params[PN_MARKER1_FREQ].dblMaxValue = pInfo->dblSampleRate / 2.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_MARKER1_FREQ].dblValue = +1000.0;
   }
  #define PN_MARKER1_LEVEL 5
  strcpy(pInfo->params[PN_MARKER1_LEVEL].sz40Name, "Marker 1 output level [dB]" );
  pInfo->params[PN_MARKER1_LEVEL].dblMinValue = -200.0;
  pInfo->params[PN_MARKER1_LEVEL].dblMaxValue = -3.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_MARKER1_LEVEL].dblValue = -50.0;
   }

  #define PN_LOCAL_OSC_FREQUENCY 6
  strcpy(pInfo->params[PN_LOCAL_OSC_FREQUENCY].sz40Name, "Shift frequency [Hz]" );
  pInfo->params[PN_LOCAL_OSC_FREQUENCY].dblMinValue = pInfo->dblAudioFreqMin; // 0 or -fs/2
  pInfo->params[PN_LOCAL_OSC_FREQUENCY].dblMaxValue = pInfo->dblSampleRate / 2.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_LOCAL_OSC_FREQUENCY].dblValue = 0.0;
   }

  // The next parameter is the frequency for a "phase-test experiment",
  //  for which the complex frequency bin is plotted in the plugin's
  //  own control window.
  //  For this, the plugin must be compiled with the option SWI_PLUGIN_WITH_GUI.
  #define PN_MARKER1_PHASE_TEST_FREQ 7
#if( SWI_PLUGIN_WITH_GUI )
  strcpy(pInfo->params[PN_MARKER1_PHASE_TEST_FREQ].sz40Name, "Phase Test frequency [Hz]" );
  pInfo->params[PN_MARKER1_PHASE_TEST_FREQ].dblMinValue = pInfo->dblAudioFreqMin; // 0 or -fs/2
  pInfo->params[PN_MARKER1_PHASE_TEST_FREQ].dblMaxValue = pInfo->dblSampleRate / 2.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_MARKER1_PHASE_TEST_FREQ].dblValue = pInfo->dblSampleRate / 4.0;
   }
#else  // ! SWI_PLUGIN_WITH_GUI ...
  strcpy(pInfo->params[PN_MARKER1_PHASE_TEST_FREQ].sz40Name, "Phase Test not available" );
#endif // SWI_PLUGIN_WITH_GUI

  // Test for the timestamps, added 2012-02-19 : 'Dirac' pulse marking a UTC second:
  #define PN_MARKER1_PPS_TEST_AMPLITUDE 8
  strcpy(pInfo->params[PN_MARKER1_PPS_TEST_AMPLITUDE].sz40Name, "PPS Test amplitude [0..1]" );
  pInfo->params[PN_MARKER1_PPS_TEST_AMPLITUDE].dblMinValue = 0.0;
  pInfo->params[PN_MARKER1_PPS_TEST_AMPLITUDE].dblMaxValue = 1.0;
  if( fRestoreDefaults )
   { pInfo->params[PN_MARKER1_PPS_TEST_AMPLITUDE].dblValue = 0.0;
   }



} // end FFP_InitParams()


//---------------------------------------------------------------------
// Helper functions for the control panel
//---------------------------------------------------------------------

#if (SWI_PLUGIN_WITH_GUI)
//---------------------------------------------------------------------
void ClearScreen( T_FftFilterInfo *pInfo )  // clears the client area
{
 HDC  hdc;
 RECT rct;

  if( pInfo->dwhPluginWindow != 0 )
   { // If the plugin's own control window exists, clear the screen:
     hdc = GetDC( (HWND)pInfo->dwhPluginWindow );
     if( hdc != NULL )
      {
        GetClientRect( (HWND)pInfo->dwhPluginWindow, &rct );
        FillRect( hdc, &rct, GetStockObject(WHITE_BRUSH) );
        ReleaseDC( (HWND)pInfo->dwhPluginWindow, hdc ); // never forget !
      } // end if < valid device context for drawing >
   }
} // end ClearScreen()
#endif // SWI_PLUGIN_WITH_GUI

#if (SWI_PLUGIN_WITH_GUI)
//---------------------------------------------------------------------
int GetTextWidth( HDC hdc, char *str )  // yet another little GDI helper..
   // returns the width of a text string, in pixels, if printed to a DC .
{
   SIZE w32Size;  // "SIZE" structure from some fancy windows header
   if( GetTextExtentPoint32( hdc, str, strlen(str), &w32Size ) )
    { return w32Size.cx;
    }
   else // GetTextExtentPoint32() failed for some reason..
    { return 0;
    }
} // end GetTextWidth()
#endif // SWI_PLUGIN_WITH_GUI

#if (SWI_PLUGIN_WITH_GUI)
//---------------------------------------------------------------------
void PrintToDC( HDC hdc, int x, int y, char *str )
  // Prints a null-terminated C-string to the graphic device context .
  // x or y < 0 means "don't change, use current position from previous call"
{
  static int cursx = 0;
  static int cursy = 0;
  SIZE tsize;
  int slen=strlen(str);
  if(x<0) x=cursx;
  if(y<0) y=cursy;
  TextOut( hdc, x, y, str, slen );
       // Note: if DevCpp throws an error like
       //   "[Linker error] undefined reference to 'TextOutA@20',
       //  add the "libgdi32.a" library to the project !
       // DevCpp doesn't include that lib automatically
       //  because the project started as a "Win32 DLL",
       //  not a "Win32 GUI Application" . This may apply to
       //  other linker errors too, so DON'T REMOVE THIS COMMENT .
  cursx = x;
  cursy = y;
  if( GetTextExtentPoint( hdc, str, slen, &tsize ) )
   { cursx += tsize.cx;
   }
} // end Print2DC()
#endif // SWI_PLUGIN_WITH_GUI



//---------------------------------------------------------------------
// Optional event handlers, called from FFP_framework :
//---------------------------------------------------------------------
#if (SWI_PLUGIN_WITH_GUI)
//---------------------------------------------------------------------
int FFP_OnKeyDown( T_FftFilterInfo *pInfo, int nVirtKey, long lKeyData )
  // Called when a key was pressed in the plugin's own window (WM_KEYDOWN) .
{

  pInfo=pInfo;          // to suppress 'unused'-warnings
  nVirtKey=nVirtKey;
  lKeyData=lKeyData;

  switch( nVirtKey )    // often upper-case ASCII, or VK_... from WINUSER.H
   {
     case 'C':  // clear the plot-screen
        ClearScreen( pInfo );
        break; // end case 'c'
     default:
        break;
   } // end switch(key)

  return 0;  // return zero if application processes the WM_KEYDOWN message .
} // end FFP_OnKeyDown()
#endif // SWI_PLUGIN_WITH_GUI



//---------------------------------------------------------------------
// Exported DLL functions
//---------------------------------------------------------------------

//---------------------------------------------------------------------
API_EXPORT(int) FFP_Init(T_FftFilterInfo *pInfo)
  /* Initializes the Filter Plugin (module prefix FFP = Fft Filter Plugin).
   * Called ONCE(!) for every instance of a plugin . See also: FFP_Exit() .
   *  Input: pInfo = pointer to an "info structure" initialized by the host.
   *                 Some members which are valid before calling FFP_Init:
   *    pInfo->dblSampleRate   : samples per second in the time domain
   *    pInfo->iNumFrequencyBins: number of complex frequency bins .
   *                 FFP_Init may need to know this to allocate some
   *                 internal buffers before we start processing.
   *  Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   *
   */
{
  BOOL fRestoreDefaults = FALSE;
  int  i;

  // Compatibility check.. see ..\fft_filter_plugins\readme.txt !
  i = sizeof(T_FftFilterInfo);  // place in a variable to see it in the debugger..
  if( pInfo->iSizeOfStruct != i )
   { // big problem: The plugin is INCOMPATIBLE with Spectrum Lab,
     //  because the size of the plugin's T_FftFilterInfo structure
     //  doesnt match what the host expects -> refuse to work !
     return FFP_ERROR_VERSION_INCOMPATIBLE;
   }

  // Fill some fields of the plugin info structure:
  strcpy(pInfo->sz255InfoText,"Adds Range 1 to Range 2 (=destination)");

  // If the filter plugin supports parameters which shall be visible
  // in SpecLab's filter control panel (on the "Plugin" tab),
  // define those parameters here, including name and value range .
  // Note : SpecLab will now save pInfo->params[].dblValue between
  //        two sessions in SETTINGS.INI, or in a *.USR file;
  //        so it's up to the plugin to overwrite these values here OR NOT.
  //        Suggestion: Check if the values are "valid", and if not,
  //        leave them as they are.   Otherwise, restore the defaults.
  if( pInfo->dwInitFlags & FFP_IFLAG_RESTORE_DEFAULTS )
   { // The host asks us to restore the default values,
     //  even if the parameters currently look "valid" .
     // (happens if the user has clicked "Restore Defaults" in the menu)
     fRestoreDefaults = TRUE;
   }
  // There may be other reasons (which only the plugin can know)
  //  which require to restore the default parameters :
  if( (pInfo->params[PN_SOURCE_START].dblMaxValue < (-pInfo->dblSampleRate/2.0) )
    ||(pInfo->params[PN_SOURCE_START].dblMaxValue > ( pInfo->dblSampleRate/2.0) )
    )
   { fRestoreDefaults = TRUE;
   }

  FFP_InitParams(pInfo, fRestoreDefaults );

  // FFP_PrintErrorString( pInfo, "Plugin compiled %s", __DATE__ );

  // Allocate an "extra" part which we need to pass to all subroutines
  //  (which would be METHODS in a C++ world; and pInfo would be 'this') :
  pInfo->pInstData = (void*)malloc( sizeof(T_MyInstData) );
  if( pInfo->pInstData == NULL )
   { return FFP_ERROR_OUT_OF_MEMORY; // can this really happen under windoze ?
   }
  // Initialize date for "this instance" : THIS_INST = pInfo->pInstData .
  THIS_INST->dblPhasePlotRe = 0.0; // real part for phase-plot, normalized to +/- 1.0
  THIS_INST->dblPhasePlotIm = 0.0; // imaginary part "  " " ..
  THIS_INST->old_phase_x    = -1;  // no "old" vector to be removed from plot-screen
  THIS_INST->fDoPhasePlot = FALSE; // flag from audio thread for plotter
  THIS_INST->dblNcoPhase = 0.0;    // initial phase value for "local oscillator"

#if(SWI_PLUGIN_WITH_GUI)  // OPTIONAL : Plugin with it's own, small control panel.
  // Set a unique "window class name" for the plugin.
  // Used in FFP_OpenControlPanel() when registering the window class (RegisterClassEx).
  strcpy(pInfo->sz40WinClassName, C_PLUGIN_WINDOW_CLASS ); // MUST BE UNIQUE !
  // The text which appears in the window title contains the CHANNEL NUMBER :
  sprintf( THIS_INST->sz80WindowTitle, "FFT Filter Plugin Demo [%d]",
           (int)pInfo->iChannel+1 );
  FFP_OpenControlPanel( pInfo, THIS_INST->sz80WindowTitle );
#endif // SWI_PLUGIN_WITH_GUI


  return FFP_ERROR_ALL_OK;
} // end FFP_Init()


//--------------------------------------------------------------------------
API_EXPORT(void) FFP_ProcessTimeDomain2(
  T_FftFilterInfo *pInfo, // [in]: pointer to structure with "additional" info;
                          //  pInfo->dwSeconds & dwNanoseconds are the timestamp
                          //  (Unix format) of the 1st sample in the array below
  int called_when,    // [in] WHEN this function is called.. one of the following:
    // FFP_PTD_CALLED_BEFORE_FORWARD_FFT : called directly before the forward FFT
    // FFP_PTD_CALLED_AFTER_INVERSE_FFT  : called directly after the inverse FFT
    //      Other callers, outside the FFT-based filter, where the
    //      number of samples is NOT related to the FFT blocksize:
    // FFP_PTD_CALLED_BEFORE_DSP_1  : in SL's circuit: when passing "L1"
    // FFP_PTD_CALLED_BEFORE_MIXER  : in SL's circuit: when passing "L2"
    // FFP_PTD_CALLED_BEFORE_FILTER : in SL's circuit: when passing "L3"
    // FFP_PTD_CALLED_BEFORE_DSP_2  : in SL's circuit: when passing "L4"
    // FFP_PTD_CALLED_BEFORE_OUTPUT : in SL's circuit: when passing "L5"
  float *pfltSource1,     // [in,out]: real part, or 1st channel, in the time domain
  float *pfltSource2,     // [in,out]: optional imaginary part, or 2nd channel (may be NULL)
  int   iNumSamples)      // [in]: number of sample points in the time domain
  /* Called by the high-priority audio thread
   *  to process the next block of samples IN THE TIME DOMAIN .
   *  This function is optional, which means it is not expected
   *  to exist in the plugin DLL (only required for "special applications").
   *  Spectrum Lab calls this function a couple of times at different places
   *  in the 'test circuit, as indicated by parameter 'called_when' .
   *  So the plugin can decide at which point to 'tap' the data .
   *  Also see notes on 'what to avoid' in FFP_ProcessSpectrum() .
   */
{
  int i, iTabIndex;
  double nco_phase = THIS_INST->dblNcoPhase; // use local copy for speed
  double phz_inc;
  long double tstamp = // Timestamp (number of seconds since 1970-01-01 00:00:00 UTC)
      (long double)pInfo->dwSeconds + 1e-9 * (long double)pInfo->dwNanoseconds;
  float *pfltCos = pInfo->pfltCosTable; // pointer to a global cosine table (one period with 32768 points)

  // Is this the point in the circuit where this plugin wants to tap the signal ?
  if( called_when != FFP_PTD_CALLED_BEFORE_FORWARD_FFT )
   { return; // not the wanted tap in the circuit !
   }

  // Inject a 'dirac' (shortest possible pulse) at the begin of a UTC second ?
  if( pInfo->params[PN_MARKER1_PPS_TEST_AMPLITUDE].dblValue != 0.0 )
   {
     // Calculate the array index of the sample closest to the next UTC second.
     // For example, if tstamp is XXXX.9000 [seconds], the next 'diract'
     // will be at index < 0.1 seconds * SampleRate > :
     i = (int) ( 0.5 + ( 1.0 - fmodl( tstamp, 1.0 ) ) * pInfo->dblSampleRate);
     if( i>=0 && i<iNumSamples )
      { // The sample with 'one-second-tick' is contained in this block of samples:
        pfltSource1[i] += pInfo->params[PN_MARKER1_PPS_TEST_AMPLITUDE].dblValue;
        if( pfltSource2 )  // same for the 2nd channel :
         { pfltSource2[i] += pInfo->params[PN_MARKER1_PPS_TEST_AMPLITUDE].dblValue;
         }
        // Note: One block of sample contains less than one second of data.
        //       Otherwise, we'd have to do this in a loop !
      }
   }


  // Perform a frequency-shift implemented in the TIME domain (multiply with NCO) ?
  if( pInfo->params[PN_LOCAL_OSC_FREQUENCY].dblValue != 0 )
   {
# define L_USE_COSINE_TABLE 1
# if( ! L_USE_COSINE_TABLE )  // here: generate sin+cos without a table..
     phz_inc = pInfo->params[PN_LOCAL_OSC_FREQUENCY].dblValue
              * 2.0 * C_PI / pInfo->dblSampleRate;
     if( (pfltSource1!=NULL) && (pfltSource2!=NULL) )
      { // If BOTH input channels are used, assume they contain a complex signal
        // ("I/Q stream") .
        // Here: multiply each complex sample with the complex output
        //       of an NCO (numerically controlled oscillator) .
        for( i=0; i<iNumSamples; ++i)
         { pfltSource1[i] *= cos( nco_phase ); // channel 1 multiplied with NCO's "I"-output
           pfltSource2[i] *= sin( nco_phase ); // channel 2 multiplied with NCO's "Q"-output
           // Note: Some books define the Q-signal as "-sin(omega*t)". We don't.
           // A POSITIVE oscillator frequency moves the input "higher" in frequency.
           nco_phase += phz_inc;
           if( nco_phase >  2*C_PI )
               nco_phase -= 2*C_PI;
           if( nco_phase < -2*C_PI )
               nco_phase += 2*C_PI;
         }
      } // end if < both channels used >
#   else   //  L_USE_COSINE_TABLE
     // Same 'frequency shifting' algorithm as above,
     //  but here using the large sine-table instead of calculating sin+cos .
     // This works a bit like a DDS (direct digital synthesizer) .
     phz_inc = pInfo->params[PN_LOCAL_OSC_FREQUENCY].dblValue
               * 32768.0 / pInfo->dblSampleRate;  // frequency -> DDS stepwidth
     if( (pfltSource1!=NULL) && (pfltSource2!=NULL) )
      { // If BOTH input channels are used, assume they contain a complex signal
        // ("I/Q stream") .
        // Here: multiply each complex sample with the complex output
        //       of an NCO (numerically controlled oscillator) .
        for( i=0; i<iNumSamples; ++i)
         {
           while( nco_phase >  32768.0 ) // circular wrap for the oscillator phase...
                  nco_phase -= 32768.0;
           while( nco_phase <  0 )       // ... in both directions,
                  nco_phase += 32768.0;  // because negative frequencies are legal
           iTabIndex = (int)nco_phase;   // index into COSINE-table
           pfltSource1[i] *= pfltCos[iTabIndex]; // channel 1 multiplied with NCO's "I"-output
           // sin(x) = cos(x-pi/2) = cos(x+3*pi/2)
           iTabIndex = (iTabIndex+24576) & 32767; // -> index to read SIN from COS-table
           pfltSource2[i] *= pfltCos[iTabIndex];  // channel 2 multiplied with NCO's "Q"-output
           // Note: Some books define the Q-signal as "-sin(omega*t)". We don't.
           // A POSITIVE oscillator frequency moves the input "higher" in frequency.
           nco_phase += phz_inc;
        }
     } // end if < both channels used >
#   endif  // L_USE_COSINE_TABLE
   } // end if( pInfo->params[PN_LOCAL_OSC_FREQUENCY].dblValue != 0 )


  THIS_INST->dblNcoPhase = nco_phase;

} // end FFP_ProcessTimeDomain()


//---------------------------------------------------------------------
API_EXPORT(int) FFP_ProcessSpectrum(
  T_FftFilterInfo *pInfo,    // [in]: pointer to structure with "additional" info
  float *pfltFreqBinsRe,     // [in+out]: complex frequency bins, REAL part
  float *pfltFreqBinsIm,     // [in+out]: complex frequency bins, IMAGINARY part
  int   iNumFrequBins )      // [in]: number of frequency bins to be processed
  /* Called by the high-priority audio thread to process
   *  the next block of samples IN THE FREQUENCY DOMAIN .
   * Note: "What to avoid" ...
   *  - Unlike FFP_Init, FFP_Exit, and FFP_OnTimer,
   *    this function (FFP_ProcessSpectrum) must not call
   *    any windows API functions - especially no GDI stuff,
   *    because many of those functions are definitely
   *    not thread-safe (reentrant).
   *  - This function should return as fast as possible.
   *    Don't use any blocking calls in here, even if they are
   *    thread-safe (like waiting for a semaphore, etc) !
   *    If this function eats up more than a few dozen milliseconds,
   *    it may cause annoying 'dropouts' in the audio stream .
   *
   *  Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   */
{
 int i,i1,i2,i3, i_max;
 float amp,gain;

  // A special experiment: Graphic display of a certain complex frequency bin,
  //   using the plugin's own little control window .
  // This 'sample' for the phase meter is taken before any further step .
#ifdef PN_MARKER1_PHASE_TEST_FREQ
  i = FFP_FrequencyToBinIndex(pInfo, pInfo->params[PN_MARKER1_PHASE_TEST_FREQ].dblValue);
  if( i>=0 && i<iNumFrequBins && !THIS_INST->fDoPhasePlot )
   { // Copy the complex frequency bin, normalized to +/- 1.0,
     // into a variable so it can be plotted later (in FFP_OnTimer) :
     THIS_INST->dblPhasePlotRe = pfltFreqBinsRe[i] / pInfo->dblMaxBinAmpl;
     THIS_INST->dblPhasePlotIm = pfltFreqBinsIm[i] / pInfo->dblMaxBinAmpl;
     if( (i&1)!=0 && (pInfo->dwFftCounter & 1)!=0)
      { // because the angle in EVEN frequency bins jumps by 180
        // for every 2nd FFT (caused by the 50% overlap) :
        THIS_INST->dblPhasePlotRe = -THIS_INST->dblPhasePlotRe;
        THIS_INST->dblPhasePlotIm = -THIS_INST->dblPhasePlotIm;
      }
     THIS_INST->fDoPhasePlot = TRUE;
   }
#endif // defined PN_MARKER1_PHASE_TEST_FREQ

  // Main purpose of this plugin:  add or subtract one frequency range from another
  i_max = pInfo->iNumFrequencyBins-1;
  if( i_max>=iNumFrequBins )
      i_max= iNumFrequBins-1;
  i1 = FFP_FrequencyToBinIndex(pInfo, // source range start index
         pInfo->params[PN_SOURCE_START].dblValue );
  i2 = FFP_FrequencyToBinIndex(pInfo, // source range end index
         pInfo->params[PN_SOURCE_START].dblValue
       + pInfo->params[PN_RANGE_WIDTH].dblValue );
  i3 = FFP_FrequencyToBinIndex(pInfo, // destination range start index
         pInfo->params[PN_DESTINATION_START].dblValue );
  gain = pInfo->params[PN_GAIN_FACTOR].dblValue; // gain for source range

  // For odd index-differences, invert the sign in every 2nd FFT .
  // (this is due to 50% overlap, see explanation in FFP_ShiftFrequencies)
  if( ( (i3-i2) & 1)!=0 && (pInfo->dwFftCounter & 1)!=0)
   { gain = -gain;
   }

  // Because the source- and destination range may overlap,
  // take care that the "inputs" dont get overwritten before being processed.
  // Furthermore, the *source* range width may be NEGATIVE to mirror frequencies.
  if( i2>=i1 )  // no mirror...
   { if( i3<=i1 )  // destination index range below source index range:
      { // use INCREASING array index
        //     i1   i2
        //    [source]
        // [destin]
        //  i3--->
        for(i=i1; (i<=i2) && (i3<i_max); ++i)
         {
           pfltFreqBinsRe[i3] += gain * pfltFreqBinsRe[i];
           pfltFreqBinsIm[i3] += gain * pfltFreqBinsIm[i];
           ++i3;
         }
      }
     else // ( i3 > i1 ) , i.e. destination index range above source index range:
      { // use DECREASING array index !
        //     i1   i2
        //    [source]
        //       [destin]
        //        <---i3
        i3 += (i2-i1);
        for(i=i2; (i>=i1) && (i3>=0); --i)
         {
           pfltFreqBinsRe[i3] += gain * pfltFreqBinsRe[i];
           pfltFreqBinsIm[i3] += gain * pfltFreqBinsIm[i];
           --i3;
         }
      }
   }
  else // i2<i1, which means the SOURCE RANGE shall be mirrored before adding !
   {

   }


  // If the "marker" is enabled (output level > -200 dB), produce marker tone.
  amp = pInfo->params[PN_MARKER1_LEVEL].dblValue;  // marker tone level [dB]
  if( amp > -200 )
   { i = FFP_FrequencyToBinIndex(pInfo, pInfo->params[PN_MARKER1_FREQ].dblValue);
     // Convert output level ["dB below the clipping point"] into a linear amplitude:
     amp = pInfo->dblMaxBinAmpl * pow( 10.0, amp / 20.0 ) ;
     // Because of the 50% FFT overlap, the phase in ODD FFT BINS
     // must be inverted (180 shift) in every 2nd calculation :
     if( i & pInfo->dwFftCounter & 1) amp = -amp;
     pfltFreqBinsRe[i] += amp;   // add the "marker tone" to the FFT
     // Note: Due to the cos^2-windowing, it is not completely correct
     //       to place the tone in ONE frequency bin only, because
     //       the cos^2-windowing widens the effective bin-width,
     //       so theoretically the neighbour bins should get their share, too.
     //       Look at the FFT of a "pure tone" - it's not just a single peak !
     // But here, we don't care, the marker tone is clean enough
     // after the inverse FFT .
   } // end if < produce "marker"-tone ? >


  return FFP_ERROR_ALL_OK;
} // end  FFP_ProcessSamples()

//--------------------------------------------------------------------------
static int FrequencyToBinIndex(
     double dblFreq_Hz,         // [in] frequency in Hertz
     double dblBinWidthInHz,    // [in] width of a single frequency bin in Hertz
     double dblAudioFrequencyOffset_Hz, // [in] audio center frequency of 1st bin
     int    iComponentsPerBin )
  // converts a frequency (in Hertz) into a "FFT frequency bin index"
{
  int iResult = 0;
  if( dblBinWidthInHz > 0 )
    iResult = (int)( 0.5 + ( dblFreq_Hz - dblAudioFrequencyOffset_Hz ) / dblBinWidthInHz);
  if(iResult<0)
     iResult=0;  // avoid negative array indices (for simplicity!)  
  return iResult * iComponentsPerBin;  
} // end FrequencyToBinIndex()
          

//--------------------------------------------------------------------------
API_EXPORT(int) FFP_ProcessDisplaySpectrum(T_FftFilterInfo *pInfo,
     float *pfltFreqBins,   // FFT frequency bins from main spectrum analyser (REAL or COMPLEX!)
     int    iNumFrequBins,  // number of frequency bins in *pfltFreqBins
     int    iFftOutputType, // 0=real(amplitudes only), 1=radio direction finder, 2=complex spectrum
     int    iAnalyserChannel, // 0=1st channel, 1=second channel
     double dblBinWidthInHz, // width of a single frequency bin in Hertz
     double dblAudioFrequencyOffset_Hz ) // audio center frequency of 1st bin
  /* Called from Spectrum Lab's DISPLAY THREAD (not from the audio worker thread).
   * This optional function in a FILTER PLUGIN(!) may take a look
   * at the FFT calculated for the *main frequency analyser*,
   * and even modifiy some of these bins before they are buffered and displayed.
   * This can be used, for example, to compensate a filter's frequency response
   * (especially at the edges), etc  etc .
   *  Caution, this function in the FFT Filter Plugin is called
   * from a different thread (not from the thread which calls ProcessSamples().
   * Use thread synchronisation objects internally if necessary !
   */
{
  int i,iFreqBinIndex;
  int iComponentsPerBin = 0;

  (void)pInfo;  // not used yet (IN THIS SIMPLE DEMO)
  (void)iAnalyserChannel; // also unused (here)

  switch( iFftOutputType )   // can we handle the data type ... ?
   {    
     case FFT_OUT_TYPE_AMPL_ONLY: /* amplitude spectrum; one "float" per frequency bin */
        iComponentsPerBin = 1;
        break;  // .. yes, we can :)
     case FFT_OUT_TYPE_RADIO_DF : /* radio direction finder spectrum, two floats/bin */
        break;  // radio direction finder mode not supported here
     case FFT_OUT_TYPE_COMPLEX  : /* generic complex spectrum, two floats per bin */
        // This is usually the case if the input is taken from an image-rejecting
        // direct conversion receiver, or a software-defined radio with quadrature
        // output like SDR-IQ and Perseus !
        iComponentsPerBin = 2;
        break;  // be sure to support this, too !
   }
  if( (iComponentsPerBin <= 0) || (dblBinWidthInHz<=0.0) )
   {  return FALSE;  // spectrum data type not supported, or illegal parameters
   } 
  // Quite a nonsense-demo but it explains the principle:
  // Ten discete frequencies (100,200,...1000 Hz) are set to ZERO for the display,
  // and ten other frequencies (1100,1200,..2000 Hz) are set to ONE (=max=0dB) for the display.
  for(i=1;i<=20;++i)
   { 
     iFreqBinIndex = FrequencyToBinIndex( 100.0/*Hz*/ * i, 
                 dblBinWidthInHz,dblAudioFrequencyOffset_Hz,iComponentsPerBin );
     if(iFreqBinIndex<iNumFrequBins)
      { if( i<=10 )
         { pfltFreqBins[iFreqBinIndex] = 0.0;  // set this frequency bin to zero
         }
        else 
         { pfltFreqBins[iFreqBinIndex] = 1.0;  // set this frequency bin to ONE = max = 0 dBfs
         }
        if( iComponentsPerBin > 1 )
         { pfltFreqBins[iFreqBinIndex+1] = 0.0;  // set the imaginary part to zero, too
         }
      }
   }

  return TRUE;

} // end FFP_ProcessDisplaySpectrum()


//--------------------------------------------------------------------------
API_EXPORT(void) FFP_OnDrawSpectrum( T_FftFilterInfo *pInfo,
     T_FFP_SPECTRUM_DRAWING_INFO *pDrawingInfo, int called_when )
  /* Called from Spectrum Lab's DISPLAY THREAD (not from the audio worker thread)
   * before(*), while(*), and after(*) the Spectrum Graph in the main window
   * is painted. This may be used to paint additional information into
   * the main window (by the filter plugin).
   * (*) The parameter 'called_when' tells the plugin when this function
   *     is being called at the moment, using one of the values defined below.
   *     This will give the plugin at least a bit of control over the Z-order.
   */
{
  HDC  hdc; // windows 'device context'
  HPEN hPen,hOldPen;
  int x,y,text_width,xt;
  char *cp;
  DWORD dwColor,dwOldTextColor,dwOldBkgndColor;

  // If valid, update some "calibration" parameters for FFP_framework1.c :
  if( pInfo->dblAdcMaxInputVoltage > 0.0 )
   { // max ADC input voltage [Volts] at the clipping point
     FFP_dblAdcMaxInputVoltage = pInfo->dblAdcMaxInputVoltage;
   }
  if( pInfo->dblAdcInputImpedance > 0.0 )
   { // input impedance [Ohms] for some exotic VOLTAGE / POWER conversions :
     FFP_dblAdcInputImpedance = pInfo->dblAdcInputImpedance; // usually 50 Ohms,
     // but this may have been changed by the user on the 'system settings' tab.
   }

  // Time to do some our own "painting" in Spectrum Lab's main spectrum display?
  hdc = (HDC)pDrawingInfo->dwhDevContext; // 'device context' for GDI functions
  switch( called_when )
   {
     case FFP_DRAW_CALLED_BEFORE_PAINTING_GRID :
        /* called between erasing the background and painting the grid    */
        PrintToDC( hdc, 8/*x*/, 0/*y*/, "Plugin: background" );
        break;
     case FFP_DRAW_CALLED_BEFORE_PAINTING_RADIO_STATIONS:
        /* called between painting the grid and the RADIO STATION INFO */
        break;
     case FFP_DRAW_CALLED_BEFORE_PAINTING_AUX_CURVES  :
        /* called between painting the radio station info and drawing the 'auxiliary' curves */
        break;
     case FFP_DRAW_CALLED_BEFORE_PAINTING_MAIN_CURVES :
        /* called between painting the 'auxiliary' and the 'main' curves */
        // For the simple "demo", show an indicator for one the filter plugin's parameters:
        dwColor = pDrawingInfo->dwSpectrumPenColor[FFP_SPECTRUM_PEN_USER1];
        hPen = CreatePen( PS_SOLID, 0/*single pixel wide*/, dwColor );
        hOldPen = SelectObject( hdc, hPen );
        dwOldTextColor = (DWORD)SetTextColor( hdc, dwColor );
        dwOldBkgndColor= (DWORD)SetBkColor( hdc, pDrawingInfo->dwSpectrumBackgroundColor );
        if( FFP_DrawSpectrum_FreqAndAmplitudeToScreenPos(
             pDrawingInfo, // [in] info about SL's main spectrum display
             pInfo->params[PN_MARKER1_FREQ].dblValue,  // [in] baseband frequency in Hertz
             pInfo->params[PN_MARKER1_LEVEL].dblValue, // [in] amplitude, magnitude or power..
             SCALE_UNIT_dBfs, // [in] scaling unit for 'amplitude, magnitude or power', see fft_filter_dll.h
             &x, &y ) )       // [out] pixel coordinate for painting and printing
         { // The pixel coordinate seems to be inside the "curve plotting area".
           // Note: Don't expect 'x' to run along the frequency axis !
           //       FFT_SpectrumGraph_FreqAndAmplitudeToScreenPos() takes care about
           //       swapping 'x' and 'y' depending on pDrawingInfo->iOrientation ,
           //       but it's your decision to rotate the text by 90 degrees for output.
           Ellipse( hdc, x-2, y-2, x+3, y+3 ); // Win32 API function, uses the current pen
           MoveToEx( hdc, x, y, NULL );
           cp = "Plugin-Marker1";
           text_width = GetTextWidth( hdc, cp );
           if( pDrawingInfo->iOrientation==0 ) // normal (horizontal) frequency axis
            { LineTo( hdc, x, pDrawingInfo->iCurveAreaY2-16 );
              xt = x-text_width/2;
              if(xt<0) xt=0;
              PrintToDC( hdc, xt, pDrawingInfo->iCurveAreaY2-16, cp );
            }
           else // rotated display, vertical frequency axis
            { LineTo( hdc, pDrawingInfo->iCurveAreaX1, y );
              PrintToDC( hdc, pDrawingInfo->iCurveAreaX1, y, cp );
            }
         } // end if < amplitude/frequency coordinate for the marker within the visible range >
        SetBkColor(   hdc, (COLORREF)dwOldBkgndColor );
        SetTextColor( hdc, (COLORREF)dwOldTextColor );
        SelectObject( hdc, hOldPen );   // never forget to restore the default pen,
        DeleteObject( hPen );  // and never forget to delete (free) a temporary GDI object
        break;
     case FFP_DRAW_CALLED_AFTER_PAINTING_CURVES  :
        /* called after drawing the curves, painting now will overwrite them */
        PrintToDC( hdc, 8/*x*/, 16/*y*/, "Plugin: foreground" );
        break;
     default:  // must be some future layer / call; ignore it here
        break;
   } // end switch( called_when )
} // end FFP_OnDrawSpectrum()

//--------------------------------------------------------------------------
API_EXPORT(void) FFP_OnTimer(T_FftFilterInfo *pInfo)
  /* Periodically called from the main thread (not from a worker thread! ) .
   * In contrast to ProcessSamples(),  OnTimer() may call windows functions,
   * do graphic stuff, etc etc.
   *
   * The calling interval is approximately 200 ms, but DON'T RELY ON THAT :
   *  The source is a windows timer event, which may be slower if the CPU
   *  is heavily loaded.
   */
{

  // .. place your own code here ..
  //  (if your plugin needs to do something periodically,
  //   thread-safe called from Spectrum Lab's main thread)
  double ampl, phase, re, im;
  int  x,y, max_x, max_y, center_x, center_y;
  char sz80Temp[84];
  HDC  hdc;           // 'device context' for windows graphic functions
  HPEN hOldPen;
  RECT rct;

  if( pInfo->dwhPluginWindow != 0 )
   { // If the plugin's own control window is visible, show some info there:
    hdc = GetDC( (HWND)pInfo->dwhPluginWindow );
    if( hdc != NULL )
     {
       GetClientRect( (HWND)pInfo->dwhPluginWindow, &rct );
       max_x = rct.right-1;
       max_y = rct.bottom-1;

       // Show FFT counter, amplitude (Am), and phase (Ph) of the "test bin"  .
       //  The complex "test bin" from the FFT (dblPhasePlotRe,dblPhasePlotIm)
       //  has already been normalized to the range +/- 1.0 .
       //  Convert it from a rectangular to a polar coordinate:
       re = THIS_INST->dblPhasePlotRe;
       im = THIS_INST->dblPhasePlotIm;
       ampl = sqrt( re*re + im*im );
       phase= (180.0/C_PI)*CalculateAngle(re,im);
       sprintf(sz80Temp, "FftCnt=%d Am=%4.1lf Ph=%5.1lf ",
          (int)pInfo->dwFftCounter, (double)ampl, (double)phase );
       PrintToDC( hdc, 0/*x*/, 0/*y*/, sz80Temp );
       sprintf(sz80Temp, "Fcb=%5.3lf Hz ",(double) /* center frequency of FFT "test"-bin */
         FFP_BinIndexToFrequency( pInfo,
           FFP_FrequencyToBinIndex( pInfo,
              pInfo->params[PN_MARKER1_PHASE_TEST_FREQ].dblValue ) ) );
       PrintToDC( hdc, -1, -1, sz80Temp );
       PrintToDC( hdc, -1, -1, "  " );

       // Add another pixel to the "phase plot" ?
       if( THIS_INST->fDoPhasePlot && max_x>100 && max_y>50)
        { // input: g_dblPhasePlotRe+g_dblPhasePlotIm, normalized to +/- 1.0
          center_x = 40 + (max_x-40) / 2;
          center_y = 10 + (max_y-10) / 2;
          x = 40 + (int)( (0.5 + re * 0.5) * (max_x-40) );
          y = 10 + (int)( (0.5 - im * 0.5) * (max_y-10) );
          if(THIS_INST->old_phase_x>=0)
           { // erase old vector ?
             hOldPen = SelectObject( hdc, GetStockObject(WHITE_PEN) );
             #define MoveTo(dc,x,y) MoveToEx(dc,x,y,NULL)
             MoveTo( hdc, center_x, center_y );
             LineTo( hdc, THIS_INST->old_phase_x, THIS_INST->old_phase_y );
             // Note: "LineTo" draws to, BUT "EXCLUDING", the end coord !
             SelectObject( hdc, hOldPen );
           }
          hOldPen = SelectObject( hdc, GetStockObject(BLACK_PEN) );
          MoveTo( hdc, center_x, center_y );
          LineTo( hdc, x, y ); // draw to, BUT "EXCLUDING", the end coord
          SetPixel( hdc, x,y, RGB(0xFF,0x00,0x00) );  // plot pixel with given color
          SelectObject( hdc, hOldPen );
          THIS_INST->old_phase_x = x;
          THIS_INST->old_phase_y = y;
          THIS_INST->fDoPhasePlot = FALSE; // flag to wait for the next phase-plot sample
          // (rather lazy, but keeps it simple.. polling every 200ms here)
        } // end if( g_fDoPhasePlot )
       PrintToDC( hdc, 0,32, "[C]lear" );
       ReleaseDC( (HWND)pInfo->dwhPluginWindow, hdc ); // never forget !
     } // end if < got a valid device context >
   } // end if < plugin's own control window exists >

} // end  FFP_OnTimer()

//--------------------------------------------------------------------------
API_EXPORT(int) FFP_ExecCmd(T_FftFilterInfo *pInfo,
     char *pszCommand,     // [in]:  command (C string)
     char *psz255Response) // [out]: optional response (C string, up to 255 chars)
  /* OPTIONAL 'command handler' inside the plugin, usable for debugging etc.
   *  Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   *  Full documentation in fft_filter_dll.h  .
   */
{
  const char *pszCmds[] = { "about", "help", "restore", "show", "test", "\0" };
  int i;
  int iResult = FFP_ERROR_ALL_OK;
  for(i=0; pszCmds[i][0]>'\0'; ++i)
   { if(stricmp(pszCommand, pszCmds[i])==0) break;
   }
  switch(i)
   { case 0:  // "about"
        // wvsprintf(psz255Response, "FFT Filter Plugin demo, compiled %s", __DATE__ );
        // wvsprintf crashed here, but sprintf worked - problems with STRINGS ?
        sprintf(psz255Response, "FFT Filter Plugin demo, compiled %s", __DATE__ );
        break;
     case 1:  // "help":
        strcpy( psz255Response, "Commands:" );
        for(i=0; pszCmds[i][0]>'\0'; ++i)
         { if(stricmp(pszCommand, pszCmds[i])==0) break;
           strcat( psz255Response, " " );
           strcat( psz255Response, pszCmds[i] );
         }
        break;
     case 2:  // "restore"
        FFP_InitParams(pInfo, TRUE/*fRestoreDefaults*/ );
        ++pInfo->iParamsModifyCnt; // let the GUI redraw the parameter table
        strcpy( psz255Response, "ok." );
        break;

     case 3:  // "show"  :  command to SHOW the window (if someone closed it)
#if(SWI_PLUGIN_WITH_GUI)  // OPTIONAL : Plugin with it's own, small control panel.
        FFP_OpenControlPanel( pInfo, THIS_INST->sz80WindowTitle );
#endif // SWI_PLUGIN_WITH_GUI
        break;

     case 4:  // "test"
        break;

     // ... add other 'interpreter commands' for your filter plugin here ...

     default:
        strcpy( psz255Response, "unknown command" );
        iResult = FFP_ERROR_FUNCTION_MISSING;
        break;
   }
  return iResult;
} // end FFP_ExecCmd()

//---------------------------------------------------------------------
API_EXPORT(int) FFP_Exit(T_FftFilterInfo *pInfo)
  /* Called once by the host when processing shall stop,
   * or before closing down. The filter plugin may free up some resources
   * which it has allocated in FFP_Init(), and -if used- close its own
   * control panel.
   *  Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   */
{

  if( pInfo->iSizeOfStruct == sizeof(T_FftFilterInfo) )
   {
     FFP_DestroyControlPanel(pInfo); // destroy the plugin's control panel (if existing)
     if( pInfo->pInstData != NULL )
      {  free( pInfo->pInstData );   // free "extra" instance data if we allocated some
         pInfo->pInstData = NULL;
      }
     return FFP_ERROR_ALL_OK;
   }
  else
   { return FFP_ERROR_VERSION_INCOMPATIBLE;
   }

} // end FFP_Exit()


#ifdef __BORLANDC__
 #pragma argsused   // Borland gripes "never used" - so what  ;-)
#endif

//---------------------------------------------------------------------
// DLL main entry point.
//   Not required by SpecLab, but possibly important for your DLL.
//
// For users of Borland C++Builder :
//   Read the notes about 'DllMain' versus 'DllEntryPoint'
//   in the automagically created 'main module' which unfortunately
//   cannot be removed from the Borland DLL project .
//---------------------------------------------------------------------

//---------------------------------------------------------------------
BOOL APIENTRY DllMain(
    HINSTANCE hInst,   // [in]: Library instance handle
    DWORD reason,      // [in]: Reason this function is being called
    LPVOID reserved )  // [in]: Not used.
{
  switch (reason)
    {
      case DLL_PROCESS_ATTACH:   // Host called 'LoadLibrary'
        ++FFPFW_iDllInstanceCount;
        break;

      case DLL_PROCESS_DETACH:   // Host called 'LoadLibrary'
        --FFPFW_iDllInstanceCount;
        break;

      case DLL_THREAD_ATTACH: // called for each new THREAD the host creates,
        // even if that thread hasn't got anything to do with this DLL.
        // For that reason, DLL_THREAD_ATTACH is totally useless here .
        break;

      case DLL_THREAD_DETACH:
        break;
    }

    /* Returns TRUE on success, FALSE on failure */
    return TRUE;
}

/* EOF < fft_filter_demo_1.c > */
