//---------------------------------------------------------------------------
// File  :  fft_filter_dll.h
// Date  :  2012-04-20  (YYYY-MM-DD)
// Author:  Wolfgang Buescher  (DL4YHF)
// Location: <DL4YHF>C:\cbproj\SpecLab\fft_filter_plugins\fft_filter_dll.h
//            (everything else must be considered an 'outdated copy')
//
// Description:
//   Header file used by all plugins for Spectrum Lab's FFT-based filter.
//   Processes signals in the FREQUENCY DOMAIN (not in the time domain).
//   Compile with Dev-Cpp or Borland C++; MS-C may work but was never tested.
//
// Last changes:
//   2012-04-20 : New parameter 'called_when' in FFP_ProcessTimeDomain2() .
//                FFP_ProcessTimeDomain() remains unchanged for compatibility.
//   2012-02-18 : Added T_FftFilterInfo.dwSeconds + dwNanoseconds (Unix timestamp)
//   2009-09-30 : Added FFP_ProcessDisplaySpectrum()
//                  and FFP_OnDrawSpectrum()  .

#ifndef _FFT_FILTER_DLL_H_
#define _FFT_FILTER_DLL_H_

#if BUILDING_DLL /* compilation switch, set to 1 (one) when BUILDING the DLL.*/
/* In their endless wisdom, C compiler authors seem to put a lot of effort */
/* in making their compilers as incompatible as possible. Some examples:   */
/*  Borland C++ :  int __stdcall __declspec(dllimport)  MyFunction(void);  */
/*  DevCpp/GNU  :  __stdcall __declspec(dllimport) int  MyFunction(void);  */
/* To avoid cluttering the code with dozens of #ifdefs, we use one macro   */
/* BEFORE and another macro AFTER the data type, so the prototype in the   */
/* DLL header file looks the same for all compilers:                       */
/*    API_EXPORT(int) MyNeatFunction(int something);                       */
# ifdef __BORLANDC__
#    define API_EXPORT(type) type __stdcall __declspec(dllexport)
# else /* not __BORLANDC__ but most likely DevCpp : */
#    define API_EXPORT(type) __stdcall __declspec(dllexport) type
# endif /* !__BORLANDC__ */
#else /* Not BUILDING_DLL (but "using" the DLL) : */
# ifdef __BORLANDC__
#    define API_EXPORT(type) type __stdcall __declspec(dllimport)
# else /* not __BORLANDC__ but most likely DevCpp : */
#    define API_EXPORT(type) __stdcall __declspec(dllimport) type
# endif /* !__BORLANDC__ */
#endif /* Not BUILDING_DLL */


//--------------------------------------------------------------------------
// Data Types and structures
//--------------------------------------------------------------------------

typedef struct // T_FFPluginParam  .
{
  char sz40Name[44];  // descriptive name of this parameter (displayed in GUI)
                      // unused parameters have an empty name .
  double dblMinValue; // min. value, filled in by the DLL, read by control panel
  double dblMaxValue; // max. value, filled in by the DLL, read by control panel
  double dblValue;    // momentary value, usually set by the control panel, read by the DLL
  double dblReserve[4];  // reserved 'double' parameters for future use
  DWORD  dwFlags;     // reserved for future use (may be some "readonly-flag", etc)
} T_FFPluginParam;

#define C_FFT_FILTER_MAX_PLUGIN_PARAMS 32

typedef struct // T_FftFilterInfo. A pointer to this structure
               //   is passed to the exported plugin functions.
{ // This structure contains most (if not everything) related to
  //    a certain instance of the FFT-based FILTER PLUGIN (!) .
  // Last modifications :
  //  2012-02-18 : added dblSampleRateDrift, dwSeconds, dwNanoseconds;
  //               using the formerly 'reserved' entries to stay compatibile.
  //  2007-03-11 : added dwhAppInstance

  int iSizeOfStruct;  // set to sizeof(T_FftFilterInfo) by the host .
      // Should be used by the plugin to check compatibility :
      // if (sizeof(T_FftFilterInfo) != pInfo->iSizeOfStruct ) then bailout..

  // Info text, will be displayed on SL's filter control panel .
  // Unlike the "error text" below, the "info text" is displayed on a static
  // text control - not in a scrolling editor window.
  //  If the plugin wants a new 'info text' to be updated on the control panel,
  //  it must FIRST print the string into sz255InfoText[] ,
  //  and then increment iInfoTextModifyCnt to let the user interface know
  //  that a new text string is now ready for displaying.
  //  This reduces problems with multithreading.
  //  The same principle is also used for other strings in T_FftFilterInfo .
  char   sz255InfoText[256]; // non-scrolling (static) info text
  int    iInfoTextModifyCnt; // increment this after modifying sz255InfoText

  // Error message, may be set by plugin if a function fails.
  //  Visible on SL's control panel in a small, scrolling, multiline editor window.
  //  (unlike the "info text" which is displayed as non-scrolling static text).
  //  Can be used for debugging too .
  char   sz255ErrorText[256]; // error string, appended to a multi-line editor window
  int    iErrorTextModifyCnt; // increment this after modifying sz255ErrorText
  int    iErrorTextCommand;   // special command to control the "error" window:
    #define FFP_ETCMD_NO_ACTION    0  /* do nothing (set by GUI after execution)*/
    #define FFP_ETCMD_CLEAR_WINDOW 1  /* clear the contents of the error window */

  // Floating point parameters (64 bit) ...
  double dblSampleRate;  // current sampling rate of the A/D converter,
         // Usually something near 11025, 22050, 44100, 48000 ... 192000 [Hz].
         // Note: Spectrum Lab may permanently update this sampling rate
         // for some very special applications . For the filter plugin: READ ONLY.
  double dblFftBinWidth;  // width of an FFT bin (in the the spectrum) [Hz].
  double dblAudioFreqMin; // frequency of FIRST FFT bin [Hz], 0 or -fs/2 .
  double dblMaxBinAmpl;   // maximum "amplitude" of a single FFT bin,
                          // just at the clipping level of the A/D-converter.
         // Calculated by the host, typically FFTsize*32768/4, because :
         // > A pure input sin wave ... A*sin(wt)... will produce an fft output
         // > peak of (N*A/4)^2  where N is FFT_SIZE.
         //  [ applies to an FFT window with an average of 0.5 ]
  double dblAdcMaxInputVoltage; // calibration parameter for absolute amplitude units
  double dblAdcInputImpedance;  // required for certain VOLTAGE / POWER conversions

  double dblSamplingRateDrift; // [Hz/second], usually zero .
  double dblProcDelayFromPlugin; // processing delay from the plugin itself
                                // (SL needs this info to adjust the timestamps
                                //  after the signal ran through the plugin)
  double dblReserved[4];  // "reserved" double-precision floats in the struct (future)

  // Integer parameters ...
  int    iNumFrequencyBins; // number of complex frequency bins (in the f-domain).
         // FFP_Init() may need to know this to allocate
         // some internal buffers before we start processing.
  int    iFftBlockOverlap; // percentage of overlap between two consecutive FFTs.
         // At the moment, the overlap is ALWAYS 50 (percent), because the FFT
         // uses a cos^2 window, and   cos(x)^2 + sin(x)^2 == 1
         // (so the 50 percent FFT overlap causes no amplitude modulation) .
  DWORD  dwFftCounter;  // increments for every FFP_ProcessSpectrum,
                        // sometimes required when odd FFTs must be sign-inverted.
  int    iChannel;      // 0=upper processing branch, 1=lower processing branch
                        //   (in spectrum lab's "circuit" window)
  int    iCallWhen;     // WHEN to call the FFT filter plugin :
                        //  0 = don't call at all ("bypass") ,
                        //  1 = before other processing steps,
                        //  2 = after other processing steps .
                        // In most cases, leave this member unchanged, and let
                        // the user decide when to call the plugin.
  DWORD  dwInitFlags;   // special "flags" for the initialisation. See FFP_Init.
    #define FFP_IFLAG_RESTORE_DEFAULTS 0x00000001  /* set params[].dblValue to defaults */
    /* bits 1...32 reserved for future use, ignore them for the moment */

  // Absolute timestamp; applies to the first sample passed in a block to FFP_ProcessTimeDomain[2]().
  DWORD  dwSeconds;     // Seconds elapsed since Januar 1st, 1970, 00:00:00.0" (in UTC)
  DWORD  dwNanoseconds; // number of nanoseconds into the current UTC second (0....999999999)
         // Phase error caused by 1 ns 'granularity' for a 96 kHz signal :
         //   360 * 1ns * 96000 Hz =: 0.03456 (acceptable ?)
  int  iReserved[5];  // "reserved" 32-bit integer parameters in the struct

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Pointers to frequently used tables, which MAY be used by the plugin .
  // For example, there is a large cosine table inside SpecLab which can be
  // used to reduce the CPU load if a plugin uses sine- and cosine functions
  // (because a simpe table lookup is still faster than CALCULATING sin & cos).
  float *pfltCosTable;  // floating point cosine table, ONE period .
     // Content: pfltCosTable[i] = cos(i * 2.0 * PI / 32768);  i=0...32767 .


  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Additional information for plugins with an own control panel (window) .
  // Note that all windows-specific types like 'HINSTANCE', 'HWND', etc
  // are replaced by simple DWORDs here to avoid the necessity of including
  // dozens of windows-specific headers before including fft_filter_dll.h :
  DWORD dwhAppInstance;  // HINSTANCE to application instance (SpecLab), set by HOST
  DWORD dwhMainWindow;   // HWND to the application's main window (SpecLab), set by HOST
  DWORD dwhPluginDll;    // HMODULE returned by LoadLibrary() when loading the plugin
  DWORD dwhPluginWindow; // handle to the plugin's own window, set by PLUGIN if used
  char  sz40WinClassName[44]; // unique "class name" of the plugin's own window, set by PLUGIN
  // Position and size of the plugin's window is initialized by the host,
  // and saved in SL's configuration file. But the plugin may overwrite this:
  int iPluginWindowX, iPluginWindowY, iPluginWindowWidth, iPluginWindowHeight;


  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Adjustable 'parameters' for the FFT Filter Plugin. Can be set on the filter
  // control panel. Unused parameters have an EMPTY name ( params[x].sz16Name[0]=0 ).
  // More info about when to initialize these parameters: see FFP_Init() !
  T_FFPluginParam params[C_FFT_FILTER_MAX_PLUGIN_PARAMS];
  int iParamsModifyCnt; // incremented by the PLUGIN (!)
     // whenever the parameter-array params[] was modified during normal operation.
     // Spectrum Lab's filter control panel holds an internal copy of this
     // counter, and whenever iParamsModifyCount was changed, it will redraw
     // the string grid with the parameter values .

  void *pInstData; // Pointer to a plugin-dependent "instance" data structure .
     // Simplifies the task of adding "own" data (see fft_filter_demo_1.c) .
     // T_FftFilterInfo::pInstData is initialized with NULL before calling
     // FFP_Init().  If the plugin allocates memory for an own "extension"
     // structure, it is responsible to clean up this memory in FFP_Exit(),
     // and set pInstData=NULL to indicate it hasn't forgotten to free it !

} T_FftFilterInfo;

typedef struct t_FFP_SPECTRUM_DRAWING_INFO
{ // Info passed to some 'OnDraw'-functions in the plugin,
  // which allows the plugin to paint something into the spectrum display
  // in Spectrum Lab's main window. See FFP_OnDrawSpectrum() .
  // For the plugin, almost everything inside this strucure is READ-ONLY !

  int iSizeOfStruct; // for 'compatibility check'. Set to sizeof(T_FFP_SPECTRUM_DRAWING_INFO) by SpecLab.

  // Note that all windows-specific types like 'HINSTANCE', 'HWND', etc
  // are replaced by simple DWORDs here to avoid the necessity of including
  // dozens of windows-specific headers before including fft_filter_dll.h :
  DWORD dwhBitmap;  // handle of (to?) the windows bitmap; the CURVE AREA is just a part of it !
     // Use this handle to call Windows API functions that requires the handle
     // of a bitmap object. Pass (HBITMAP)dwhBitmap as the bitmap handle parameter
     // to these functions.
     // The plugin must not take ownership (and destroy) the bitmap handle !
  DWORD dwhDevContext; // handle of (to?) the bitmap's 'device context',
     // retrieved from the VCL using TBitmap.Canvas.Handle .
     // Some Windows API functions may requires a handle to this thing
     // (rather than a handle to the bitmap, for what it's worth...)
     // If you need this device context, cast it into a HDC like this:
     // HDC  hdc = (HDC)pDrawingInfo->dwhDevContext; // 'device context'
  int iBitmapWidth, iBitmapHeight;  // dimension of THE ENTIRE BITMAP in pixels

  int iAnalyserChannel;   // 0=first analyser channel;  1=second channel (if active)
  int iCurveAreaX1, iCurveAreaY1, iCurveAreaX2, iCurveAreaY2; // curve drawing area
  int iOrientation;   // 0=normal=horizontal frequency scale; 1=rotated=vertical frequency scale
  int iAmplScaleUnit; // SCALE_UNIT_..., originally defined in \CBproj\SoundUtl\UTILITY1.H,
                      // typically 9 = SCALE_UNIT_dBfs .
  double dblBasebandFreqMin,dblBasebandFreqMax; // displayed FREQUENCY range (baseband, i.e. without RF-offset)
  double dblGraphAmplMin,   dblGraphAmplMax;    // displayed AMPLITUDE range
  double reserved[8]; // reserved 64-bit floating point values (for future use)

  DWORD  dwSpectrumPenColor[ 8 ];  // RGB-mixes of the "pen" colours as defined
         // on Spectrum Lab's configuration panel ("display colours") .
         // Pens [0]..[4] are used for the curves painted by Spectrum Lab,
         // pens [5]..[7] are purely application-specific :
    #define FFP_SPECTRUM_PEN_MAIN        0 /* main ("current") spectrum graph */
    #define FFP_SPECTRUM_PEN_NON_AVRG    1 /* non-averaged spectrum graph     */
    #define FFP_SPECTRUM_PEN_PEAK_HOLD   2 /* peak-holding graph              */
    #define FFP_SPECTRUM_PEN_FILTER_RESP 3 /* response of the FFT-based filter */
    #define FFP_SPECTRUM_PEN_RANGE_AVERAGE 4 /* average over a range of spectra */
    #define FFP_SPECTRUM_PEN_USER1       5 /* 1st pen reserved for plugins  */
    #define FFP_SPECTRUM_PEN_USER2       6 /* 2nd pen reserved for plugins  */
    #define FFP_SPECTRUM_PEN_USER3       7 /* 3rd pen reserved for plugins  */
  DWORD  dwSpectrumBackgroundColor;
  DWORD  dwSpectrumGridColor;


} T_FFP_SPECTRUM_DRAWING_INFO;

//--------------------------------------------------------------------------
// Error codes
//--------------------------------------------------------------------------

#define FFP_ERROR_ALL_OK 0         /* no error at all */
#define FFP_ERROR_FUNCTION_FAILED -1 /* general error which doesn't match anything below */
#define FFP_ERROR_ILLEGAL_PARAM   -2
#define FFP_ERROR_VERSION_INCOMPATIBLE -3
#define FFP_ERROR_OUT_OF_MEMORY   -4
#define FFP_ERROR_FILE_ERROR      -5 /* file not found, file not opened, etc */
#define FFP_ERROR_FUNCTION_MISSING -6 /* function not supported or missing (in DLL) */

//--------------------------------------------------------------------------
// "Scale Units", an excerpt from SoundUtl/Utility1.H :
//--------------------------------------------------------------------------
// Scaling units for T_SPECTRUM.scale_type and others, so no module prefix here.
//   A copy of the more frequently used scale units is in FFT_FILTER_DLL.H ,
//   so that plugin developers don't need to include UTILITY1.H as well .
#define SCALE_UNIT_NONE    -1
#define SCALE_UNIT_UNKNOWN  0
  // Voltages, Power units ...
#define SCALE_UNIT_NORM_V   1  // normalized ADC units (voltage), +/- 1.0  for full A/D converter input range
#define SCALE_UNIT_PERCENT  2  // linear, reference is "full scale" (100 % = ADC clipping)
#define SCALE_UNIT_VOLT     3  // voltage (V), 'calibrated' using UTL_dblAdcMaxInputVoltage
#define SCALE_UNIT_WATT     4  // power (W), 'calibrated' with UTL_dblAdcMaxInputVoltage + UTL_dblAdcInputImpedance
#define SCALE_UNIT_NORM_PWR 5  // power, normalized to 0...1 for full A/D converter input range

  // Logarithmic units, dB, relative to "something" ...
#define SCALE_UNIT_MASK_dB 0x18  // mask to detect unit "dB" in all of its flavours..
#define SCALE_UNIT_dB      0x08  // dB , unknown reference level (using UTL_dblAdcMaxInputVoltage as user-defined reference)
#define SCALE_UNIT_dBfs    0x09  // dB , reference level is  "full scale" (0 dBfs = clipping point)
#define SCALE_UNIT_dBV     0x0A  // dB , reference level is 1 volt
#define SCALE_UNIT_dBuV    0x0B  // dB , reference level is 1 microvolt rms across any impedance (dBV)
#define SCALE_UNIT_dBm     0x0C  // dB , reference level is 1 milliwatt (often across 600 ohms)
  // Other UNIT CODES, which are NO amplitudes/magnitudes ...
#define SCALE_UNIT_DEGREE  0x10
#define SCALE_UNIT_HERTZ   0x11
#define SCALE_UNIT_SECONDS 0x12


//--------------------------------------------------------------------------
// Function prototypes
//--------------------------------------------------------------------------

#ifdef __cplusplus__
extern "C" {
#endif /* __cplusplus__ */


//--------------------------------------------------------------------------
API_EXPORT(int) FFP_Init(T_FftFilterInfo *pInfo);
  /* Initializes the Filter Plugin (module prefix FFP = Fft Filter Plugin).
   * Called ONCE(!) for every instance of a plugin . See also: FFP_Exit() .
   *
   *  Input: pInfo = pointer to an "info structure" initialized by the host.
   *                 Some members which are valid before calling FFP_Init:
   *    pInfo->dblSampleRate   : samples per second in the time domain
   *    pInfo->iNumFrequencyBins: number of complex frequency bins .
   *                 FFP_Init() may need to know this to allocate some
   *                 internal buffers before we start processing.
   *  Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   *
   */

//--------------------------------------------------------------------------
API_EXPORT(int) FFP_Exit(T_FftFilterInfo *pInfo);
  /* Called once by the host (SpecLab,..) to stop processing, or when terminating.
   * The filter plugin should free up the resources allocated in FFP_Init(),
   * and -if used- close its own graphic user interface.
   * Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   */

//--------------------------------------------------------------------------
API_EXPORT(int) FFP_ProcessSpectrum(
  T_FftFilterInfo *pInfo,    // [in]: pointer to structure with "additional" info
  float *pfltFreqBinsRe,     // [in], [out]:  complex frequency bins, REAL part
  float *pfltFreqBinsIm,     // [in], [out]:  complex frequency bins, IMAGINARY part
  int   iNumFrequBins );     // [in]: number of frequency bins to be processed
  /* Called by the high-priority audio thread to process
   *  the next block of samples IN THE FREQUENCY DOMAIN .
   * Note: "What to avoid" ...
   *  - Unlike FFP_Init, FFP_Exit, and FFP_OnTimer,
   *    this function (FFP_ProcessSpectrum) must not call
   *    any windows API functions - especially no GDI stuff,
   *    because many of those functions are definitely
   *    not thread-safe (reentrant).
   *  - This function should return as fast as possible.
   *    Don't use any blocking calls in here, even if they are
   *    thread-safe (like waiting for a semaphore, etc) !
   *    If this function consumes up more than a few dozen milliseconds,
   *    it may cause annoying 'dropouts' in the audio stream .
   *
   *  Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   */


//--------------------------------------------------------------------------
API_EXPORT(void) FFP_ProcessTimeDomain2( /* optional; for processing in the time domain */
  T_FftFilterInfo *pInfo, // [in]: pointer to structure with "additional" info;
                      //  pInfo->dwSeconds & dwNanoseconds contain the timestamp
                      //     (Unix format) of the 1st sample in the array below.
  int called_when,    // [in] WHEN this function is called.. one of the following:
  #define FFP_PTD_CALLED_BEFORE_FORWARD_FFT 0 /* called directly before the forward FFT */
  #define FFP_PTD_CALLED_AFTER_INVERSE_FFT  1 /* called directly after the inverse FFT  */
                      // Other callers, outside the FFT-based filter, where the
                      //  number of samples is NOT related to the FFT blocksize:
  #define FFP_PTD_CALLED_BEFORE_DSP_1  11 /* in SL's circuit: when passing "L1" */
  #define FFP_PTD_CALLED_BEFORE_MIXER  12 /* in SL's circuit: when passing "L2" */
  #define FFP_PTD_CALLED_BEFORE_FILTER 13 /* in SL's circuit: when passing "L3" */
  #define FFP_PTD_CALLED_BEFORE_DSP_2  14 /* in SL's circuit: when passing "L4" */
  #define FFP_PTD_CALLED_BEFORE_OUTPUT 15 /* in SL's circuit: when passing "L5" */
  float *pfltSource1, // [in,out] real part, or 1st channel, in the time domain
  float *pfltSource2, // [in,out] optional imaginary part, or 2nd channel (may be NULL)
              //  Only used when both FFT-filters in Spectrum Lab
              //  are combined for I/Q processing .
  int  iNumSamples);  // [in] number of sample points in the time domain
  /* Called by the high-priority audio thread to process
   *  the next block of samples IN THE TIME DOMAIN .
   *  This function is optional, i.e. it is not expected
   *  to exist in the plugin DLL (only required for "special applications").
   *  If it exists, the DLL-host will call it, otherwise it won't .
   *  Note: The number of samples processed in each call is only
   *        HALF the FFT-size, because of the 50 percent FFT overlap .
   *        In other words, every audio sample in the time domain
   *        only runs through this subroutine ONCE, not TWICE .
   */

//--------------------------------------------------------------------------
API_EXPORT(void) FFP_ProcessTimeDomain( /* OUTDATED ! Only for compatibility ! */
  T_FftFilterInfo *pInfo,  float *pfltSource1,  float *pfltSource2, int iNumSamples);
  /* Similar as FFP_ProcessTimeDomain2(), only exists for backward compatibility.
   * In contrast to FFP_ProcessTimeDomain2(), only called ONCE,
   * before the forward FFT. Actually the same as FFP_ProcessTimeDomain2()
   * with called_when == FFP_PTD_CALLED_BEFORE_FORWARD_FFT .
   * All other parameters have the same meaning as in FFP_ProcessTimeDomain2().
   */


//--------------------------------------------------------------------------
API_EXPORT(int) FFP_ProcessDisplaySpectrum(T_FftFilterInfo *pInfo,
     float *pfltFreqBins,   // FFT frequency bins from main spectrum analyser (REAL or COMPLEX!).
                            // The max. value (for 0 dBfs) is normalized to ONE .
     int    iNumFrequBins,  // number of frequency bins in *pfltFreqBins
     int    iFftOutputType, // 0=real(amplitudes only), 1=radio direction finder, 2=complex spectrum
     int    iAnalyserChannel, // 0=1st channel, 1=second channel
     double dblBinWidthInHz, // width of a single frequency bin in Hertz
                             // (in the *main spectrum analyser*, not the filter)
     double dblAudioFrequencyOffset_Hz ); // audio center frequency of 1st bin
  /* Called from Spectrum Lab's DISPLAY THREAD (not from the audio worker thread).
   * This optional function in a FILTER PLUGIN(!) may take a look
   * at the FFT calculated for the *main frequency analyser*,
   * and even modifiy some of these bins before they are buffered and displayed.
   * This can be used, for example, to compensate a filter's frequency response
   * (especially at the edges), etc  etc .
   *  Caution, this function in the FFT Filter Plugin is called
   * from a different thread (not from the thread which calls ProcessSamples().
   * Use thread synchronisation objects internally if necessary !
   * Possible values for 'iFftOutputType' below (in FFT_FILTER_DLL.H) .
   * Return value : 0=FALSE ("nothing happened" or "cannot handle the type";
   *                1=TRUE = "ok" .
   */
  #define FFT_OUT_TYPE_AMPL_ONLY 0  /* amplitude spectrum; one "float" per frequency bin */
  #define FFT_OUT_TYPE_RADIO_DF  1  /* radio direction finder spectrum, two floats/bin */
  #define FFT_OUT_TYPE_COMPLEX   2  /* generic complex spectrum, two floats per bin */

//--------------------------------------------------------------------------
API_EXPORT(void) FFP_OnDrawSpectrum( T_FftFilterInfo *pInfo,
     T_FFP_SPECTRUM_DRAWING_INFO *pDrawingInfo, int called_when );
  /* Called from Spectrum Lab's DISPLAY THREAD (not from the audio worker thread)
   * before(*), while(*), and after(*) the Spectrum Graph in the main window
   * is painted. This may be used to paint additional information into
   * the main window (by the filter plugin).
   * (*) The parameter 'called_when' tells the plugin when this function
   *     is being called at the moment, using one of the values defined below.
   *     This will give the plugin at least a bit of control over the Z-order.
   */
  #define FFP_DRAW_CALLED_BEFORE_PAINTING_GRID   0      /* called between erasing the background and painting the grid    */
  #define FFP_DRAW_CALLED_BEFORE_PAINTING_RADIO_STATIONS 1 /* called between painting the grid and the RADIO STATION INFO */
  #define FFP_DRAW_CALLED_BEFORE_PAINTING_AUX_CURVES  2  /* called between painting the radio station info and drawing the 'auxiliary' curves */
  #define FFP_DRAW_CALLED_BEFORE_PAINTING_MAIN_CURVES 3  /* called between painting the 'auxiliary' and the 'main' curves */
  #define FFP_DRAW_CALLED_AFTER_PAINTING_CURVES  4  /* called after drawing the curves, painting now will overwrite them */

//--------------------------------------------------------------------------
API_EXPORT(void) FFP_OnTimer(T_FftFilterInfo *pInfo);
  /* Periodically called from the main thread (not from a worker thread! ) .
   * In contrast to ProcessSamples(),  OnTimer() may call windows functions,
   * do graphic stuff, etc etc.
   *
   * The calling interval is approximately 200 ms, but DON'T RELY ON THAT :
   *  The source is a windows timer event, which may be slower if the CPU
   *  is heavily loaded.
   */

//--------------------------------------------------------------------------
API_EXPORT(int) FFP_ExecCmd(T_FftFilterInfo *pInfo,
     char *pszCommand,     // [in]:  command (C string)
     char *psz255Response);// [out]: optional response (C string, up to 255 chars)
  /* OPTIONAL 'command handler' inside the plugin, usable for debugging etc.
   *  The input is a plain C string like "help", "test", "restore", etc .
   *     It can be sent from the filter control panel in SpecLab
   *     by entering a command line in the MESSAGE WINDOW on the bottom
   *     of the 'FFT Filter Plugin' tab .
   *     The format, syntax, and complexity of the commands depends
   *     on the plugin. One suggested command is "restore" to let
   *     the plugin restore the default values in pInfo->params[].dblValue .
   *
   *  Some commands (which act like "functions returning a string")
   *     may place a response string in pszResponse[0..254] .
   *     This response will then be printed into the MESSAGE WINDOW,
   *     so the message window (actually a multi-line edit control)
   *     can work like a tiny debugging console .
   *
   *  Notes:
   *   - This command will usually be called from the application's
   *     MAIN THREAD, so it may be interrupted by the AUDIO PROCESSING THREAD
   *     (from which FFP_ProcessSpectrum may be called). So, depending on the
   *     type of command, the plugin may need some kind of synchronisation
   *     to avoid inconsistences; because T_FftFilterInfo *pInfo passed to
   *     FFP_ExecCmd and FFP_ProcessSpectrum point to the same memory location.
   *
   *   - Unlike most other exported functions in the plugin, FFP_ExecCmd()
   *     is only optional, which means you don't have to implement it in your
   *     DLL .
   *
   *  Return value:  one of the FFP_ERROR_.. codes (hopefully FFP_ERROR_ALL_OK)
   */

#ifdef __cplusplus__
}
#endif /* __cplusplus__ */


#endif /* _FFT_FILTER_DLL_H_ */
